/*
** Modular Logfile Analyzer
** Copyright 2000 Jan Kneschke <jan@kneschke.de>
**
** Homepage: http://www.kneschke.de/projekte/modlogan
**

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version, and provided that the above
    copyright and permission notice is included with all distributed
    copies of this or derived software.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

**
** $Id: main.c,v 1.57 2001/09/17 10:11:19 lohgan Exp $
*/

/**
 * \mainpage ModLogAn API Documenation
 * 
 * \section Introduction
 * Add docs.
 *
 */

#include <libintl.h>
#include <locale.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <assert.h>

#include <libxml/parser.h>
#include <zlib.h>

#include "config.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "mrecord.h"
#include "mstate.h"
#include "mlocale.h"
#include "mconfig.h"
#include "mplugins.h"
#include "mdatatypes.h"
#include "datatypes/webhist/datatype.h"
#include "datatypes/state/datatype.h"

#define HISTORY_FILE_VERSION	"0.2"

int history_write(mconfig *conf, mlist *l, char *subpath) {
	char filename[255];
	gzFile *fd;
	
	sprintf(filename, "%s%s%s/mla.history.xml", 
		conf->outputdir ? conf->outputdir : ".", 
		subpath ? "/" : "",
		subpath ? subpath : "");
	
	if ((fd = gzopen(filename, "wb")) == NULL) {
		fprintf(stderr, "%s.%d: can't open %s: %s\n", __FILE__, __LINE__, filename, strerror(errno));
		return -1;
	}
	
	gzprintf(fd, "<?xml version=\"1.0\"?>\n");
	gzprintf(fd, "<!DOCTYPE history PUBLIC \"-//MLA//history\" \"mla.history.dtd\">\n");
	gzprintf(fd, "<history version=\"%s\" package=\"%s\">\n",
		HISTORY_FILE_VERSION, PACKAGE);
	
	mlist_write(fd, l);
	
	gzprintf(fd, "</history>\n");
	gzclose(fd);
	
	return 0;
}

int history_read(mconfig *conf, mlist *l, char *subpath) {
	char filename[255];
	xmlDocPtr doc = NULL;
	xmlNodePtr node = NULL;
	
	sprintf(filename, "%s%s%s/mla.history.xml", 
		conf->outputdir ? conf->outputdir : ".", 
		subpath ? "/" : "",
		subpath ? subpath : "");
	
	if ((doc = xmlParseFile(filename)) == NULL) {
		fprintf(stderr, "%s.%d: *grmpf* can't parse xml file '%s': %s\n", __FILE__, __LINE__, filename, strerror(errno));
		return -1;
	}
	
	/* the first node is the documentroot */
	node = xmlDocGetRootElement(doc);
	
	/* get the sections */
	node = node->children;
	
	while (node->type != XML_ELEMENT_NODE) {
		node = node->next;
		continue;
	}
		
	if (mdata_insert_value(&l, M_DATA_FIELDTYPE_LIST, node)) {
		fprintf(stderr, "%s.%d: insert failed for '%s'\n", __FILE__, __LINE__, node->name);
		return -1;
	}
		
	xmlFreeDoc(doc);
	
	return 0;
}

void show_version() {
	printf("%s %s\n", PACKAGE, VERSION);
}

void show_header() {
	printf("%s %s\n", PACKAGE, VERSION);
}

void show_usage() {
	show_header();
	

#ifdef HAVE_GETOPT_LONG
	printf("Options:\n");
	printf(" -c <configfile>    filename of the configfile\n");
	printf(" -h --help          this help screen\n");
	printf(" -v --version       display version\n");
#else
	printf("Options:\n");
	printf(" -c <configfile>    filename of the configfile\n");
	printf(" -h                 this help screen\n");
	printf(" -v                 display version\n");
#endif
};

int restore_internal_state_old (mconfig *conf, mlist *state_list) {
	DIR *dir;
	int splitter = 0;
	
	
	/* scan the outputdir for directories */
	if ((dir = opendir(conf->outputdir)) == NULL) {
		fprintf(stderr, "can't open the outputdir: %s - %s\n", conf->outputdir, strerror(errno));
		return -1;
	} else {
		struct dirent * dent;
		while ((dent = readdir(dir))) {
			struct stat st;
			char *dirname = NULL;
			
		/* skip '.', '..' and hidden directories */
			if (*(dent->d_name) == '.') continue;
			
			dirname = malloc(strlen(conf->outputdir)+strlen(dent->d_name)+2);
			assert(dirname);
			
			sprintf(dirname, "%s/%s", conf->outputdir, dent->d_name);
			
			if (stat(dirname, &st)) {
				fprintf(stderr, "%s.%d: can't stat the output directory (%s) : %s\n", __FILE__, __LINE__, dirname, strerror(errno));
			} else if (S_ISDIR(st.st_mode)) {
				mstate *state = NULL;
				mlist *hist = mlist_init();
				
				M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
					"reading statefile from %s\n", dent->d_name);
				
				if (history_read(conf, hist, dent->d_name)) {
					mlist_free(hist);
					hist = NULL;
				}

				if (conf->incremental) {
					state = mstate_init();
			
					if (mstate_read(conf, state, 0, 0, dent->d_name)) {
						mstate_free(state);
						state = NULL;
					}
				}
				
				if (!hist || (conf->incremental && !state)) {
					if (hist) {
						mlist_free(hist);
						hist = NULL;
					} else {
						fprintf(stderr, "%s.%d: reading history failed\n", __FILE__, __LINE__);
					}
					
					if (state) {
						mstate_free(state);
						state = NULL;
					} else if (conf->incremental) {
						fprintf(stderr, "%s.%d: reading State file failed\n", __FILE__, __LINE__);
					}
				} else {
					mdata *data = mdata_State_create(dent->d_name, state, hist);
					
					mlist_insert(state_list, data);
					
					M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
						 "state restored for %s\n", dent->d_name);
					
					splitter = 1;
				} 
			}
			
			free(dirname);
		}
		closedir(dir);
	}
	
	if (!splitter) {
		mstate *state = NULL;
		mlist *hist = mlist_init();
				
		if (history_read(conf, hist, NULL)) {
			mlist_free(hist);
			hist = NULL;
		}
				
		if (conf->incremental) {
			state = mstate_init();
	
			if (mstate_read(conf, state, 0, 0, NULL)) {
				mstate_free(state);
				state = NULL;
			} else {
				splitter = 1;
			}
		}
				
		if (!hist || !state) {
			if (hist)	mlist_free(hist);
			else fprintf(stderr, "%s.%d: reading history failed\n", __FILE__, __LINE__);
			if (state)	mstate_free(state);
			else fprintf(stderr, "%s\n", _("Reading State file failed"));
		} else {
			mdata *data = mdata_State_create("", state, hist);
			mlist_insert(state_list, data);
		} 
	}
	
	return 0;
}

int restore_internal_state (mconfig *conf, mlist *state_list) {
	char *fn;
	char buf[255];
	FILE *split_file;
	
	if (conf->outputdir == NULL) {
		return -1;
	}
	
	fn = malloc(strlen(conf->outputdir) + strlen("/modlogan.statefiles") + 1);
	strcpy(fn, conf->outputdir);
	strcat(fn, "/modlogan.statefiles");
	
	
/* fallback mode */
	if ((split_file = fopen(fn, "r")) == NULL) {
		free(fn);
		return restore_internal_state_old(conf, state_list);
	}
	
	while (fgets(buf, sizeof(buf), split_file)) {
		mstate *state = NULL;
		mlist *hist = mlist_init();
		
		/* remove the \n */
		buf[strlen(buf)-1] = '\0';
		
		M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
			"reading statefile from '%s'\n", buf);
				
		if (history_read(conf, hist, buf)) {
			fprintf(stderr, "%s.%d: hist failed ?\n", __FILE__, __LINE__);
			mlist_free(hist);
			hist = NULL;
		}

		if (conf->incremental) {
			state = mstate_init();
			
			if (mstate_read(conf, state, 0, 0, buf)) {
				mstate_free(state);
				state = NULL;
			}
		}
				
		if (!hist || (conf->incremental && !state)) {
			if (hist)	mlist_free(hist);
			else fprintf(stderr, "%s.%d: reading history failed\n", __FILE__, __LINE__);
			
			if (state)	mstate_free(state);
			else if (conf->incremental) fprintf(stderr, "%s\n", _("Reading State file failed"));
			
			state = NULL;
			hist = NULL;
		} else {
			mdata *data = mdata_State_create(buf, state, hist);
					
			mlist_insert(state_list, data);
					
			M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
				"state restored for '%s'\n", buf);
		} 
	}
	
	fclose(split_file);
	free(fn);
	
	return 0;
}

typedef struct {
	mlogrec *record;
	int rec_state;
} mgetrec;


int get_next_record (mconfig *conf, mlogrec *rec) {
	mplugin *func;
	int i;
	int state = M_RECORD_EOF;
	static mgetrec **recs = NULL;
	time_t t = -1;
	int ndx = -1;
	
	if (recs == NULL) {
		
		/* a small mem leak. fix it if you like */
		recs = malloc(sizeof(mgetrec *) * conf->plugin_count);
		for (i = 0; i < conf->plugin_count; i++) {
			recs[i] = malloc(sizeof(mgetrec));
			recs[i]->record = NULL;
			recs[i]->rec_state = M_RECORD_EOF;
		}
	}
	
	for (i = 0; i < conf->plugin_count; i++) {
		func = ((mplugin **)(conf->plugins))[i];
		conf->plugin_conf = func->config;
		
		if (func->get_next_record) {
			if (recs[i]->record == NULL) {
				recs[i]->record = mrecord_init();
				recs[i]->rec_state = func->get_next_record(conf, recs[i]->record);
			}
		
			if (t == -1 || recs[i]->record->timestamp < t) {
				t = recs[i]->record->timestamp;
				ndx = i;
			}
		}
	}
	
	if (ndx != -1) {
		mrecord_copy(rec, recs[ndx]->record);
		state = recs[ndx]->rec_state;
		mrecord_free(recs[ndx]->record);
		recs[ndx]->record = NULL;
	}

	return state;
}

int generate_monthly_output(mconfig *conf, mstate *state, const char *subpath) {
	int i;
	mplugin *func;
	for (i = 0; i < conf->plugin_count; i++) {
		func = ((mplugin **)(conf->plugins))[i];
		conf->plugin_conf = func->config;
		
		if (func->gen_report &&	func->gen_report(conf, state, subpath)) return -1;
	}

	return 0;
}
int generate_history_output(mconfig *conf, mlist *history, const char *subpath) {
	int i;
	mplugin *func;
	for (i = 0; i < conf->plugin_count; i++) {
		func = ((mplugin **)(conf->plugins))[i];
		conf->plugin_conf = func->config;

		if (func->gen_history && func->gen_history(conf, history, subpath)) return -1;
	}

	return 0;
}

void dump (mlist *state_list) {
	mlist *s = state_list;
	int i;
	
	while (s) {
		mstate_web *w;
		if (!s->data) return;
		
		w = ((mstate *)(s->data->data.state.state))->ext;
	
		for (i = 0; i < 31; i++) {
			fprintf(stderr, "dump_list (%p)->'%s': %d: %ld\n", s, s->data->key, i, w->days[i].hits);
		}
		
		s = s->next;
	}
}

int insert_record(mconfig *conf, mlist *state_list, mlogrec *record) {
	int i = 0;
	mplugin *func;
	
	for (i = 0; i < conf->plugin_count; i++) {
		func = ((mplugin **)(conf->plugins))[i];
		conf->plugin_conf = func->config;

		if (func->insert_record)
			func->insert_record(conf, state_list, record);
	}

	return 0;
}

typedef struct {
	struct timeval start;
	struct timeval stop;
	long int span;
} mtimer;

#define MTIMER_START(x)\
	gettimeofday(&(x.start), NULL);

#define MTIMER_STOP(x)\
	gettimeofday(&(x.stop), NULL);

#define MTIMER_CALC(x)\
	x.span += ((x.stop.tv_sec * 1000.0 + x.stop.tv_usec / 1000.0) - (x.start.tv_sec * 1000.0 + x.start.tv_usec / 1000.0)); \
	
#define MTIMER_GET_MSEC(x) \
	x.span
	
#define MTIMER_RESET(x) \
	x.span = 0; \
	x.start.tv_sec = x.start.tv_usec = x.stop.tv_sec = x.stop.tv_usec = 0;

int main (int argc, char **argv) {
	mconfig *conf = NULL;
	mlogrec *rec = NULL;
	mlist *state_list = NULL;
	mlist *l;
	
	int ret, l_month = -1, l_year = -1, l_hour = -1, l_day = -1, i;
	char *conf_filename = NULL;
	long lines = 0, lines_corrupt = 0, lines_skipped = 0, lines_ignored = 0;
	int first_valid_line = 0;

	struct tm *tm;
	
	time_t l_stamp = -1;
	mtimer process_timer, post_process_timer, parse_timer, setup_timer;
	
#ifdef HAVE_GETOPT_LONG	
	int option_index = 0;
	static struct option long_options[] = {
		{"help", 0, NULL, 0},
		{"version", 0, NULL, 0},
		{NULL, 0, NULL, 0}
	};
#endif
	
	MTIMER_RESET(process_timer);
	MTIMER_RESET(post_process_timer);
	MTIMER_RESET(parse_timer);
	MTIMER_RESET(setup_timer);
	
#ifdef HAVE_GETOPT_LONG	
	while ((i = getopt_long(argc,argv,"c:hv", 
				long_options, &option_index)) != -1) {
		
		/* map the long options to the short ones */
		if (i == 0) {
			switch(option_index) {
			case 0: i = 'h'; break;
			case 1: i = 'v'; break;
			default:
				show_usage();
				exit( EXIT_FAILURE);
			}
		}
#if 0
		/* for jed and it formating problems */
	}
#endif
#else
	while ((i = getopt(argc,argv,"c:hv")) != EOF) { 
#endif		
		switch(i) {
		case 'c': 
			conf_filename = optarg;
			break;
		case 'h':
			show_usage();
			exit( EXIT_SUCCESS);
		case 'v':
			show_version();
			exit( EXIT_SUCCESS);
		default:
			show_usage();
			exit( EXIT_FAILURE);
		}
	}
		
	show_header();
	init_locale(conf);
	
	MTIMER_START(setup_timer);
	
	if (!(conf = mconfig_init(conf_filename))) {
		fprintf(stderr, "%s.%d: reading configfile failed - going down hard\n", __FILE__, __LINE__);
		return -1;
	}
	
	state_list = mlist_init();
	rec = mrecord_init();

	restore_internal_state(conf, state_list);
	
	if (conf->incremental) {
		l = state_list;
		
		while (l) {
			mdata *data = l->data;
			mstate *int_state;
			if (!data) break;
			
			int_state = data->data.state.state;
			if (int_state->timestamp > l_stamp) {
				l_stamp = int_state->timestamp;
				l_year = int_state->year-1900;
				l_month = int_state->month-1;
			}
			
			l = l->next;
		}
	}
		
	if (l_stamp != -1) {
		M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
			 "Ignoring all records before %s", ctime(&l_stamp));
	}

	M_DEBUG0(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_NONE,
		"startup - finished\n");
	
	MTIMER_STOP(setup_timer);
	MTIMER_CALC(setup_timer);
	
	printf("[");

	MTIMER_START(parse_timer);
	while ((ret = get_next_record(conf, rec)) != M_RECORD_EOF) {
		MTIMER_STOP(parse_timer);
		MTIMER_CALC(parse_timer);

		MTIMER_START(process_timer);
		
		lines++;
		if (ret == M_RECORD_NO_ERROR) {
			/* HACK */
			if (rec->timestamp == 0) {
				fprintf(stderr, "%s.%d: line %ld returned no timestamp !! something strange will happen. Going down hard\n", 
					__FILE__, __LINE__,
					lines);
				return -1;
			} 
		
			if (l_month == -1) {
				tm = localtime(&(rec->timestamp));
				l_month = tm->tm_mon;
				l_year = tm->tm_year;
				l_stamp = rec->timestamp;
				l_hour = tm->tm_hour;
			}

			if (l_stamp > rec->timestamp) {
				if (first_valid_line == 0 && conf->incremental == 1) {
					lines_skipped++;
					MTIMER_START(parse_timer);
					continue;
				}
				M_DEBUG2(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_VERBOSE,
					"Out of Sequence - ignoring timestamp: %ld > %ld\n", l_stamp, rec->timestamp);
				
				rec->timestamp = l_stamp;
			}
			
			tm = localtime(&(rec->timestamp));

			if (l_day != tm->tm_mday) 
				l_day = tm->tm_mday;

			if (conf->gen_report_threshold > 0 && 
				(lines - lines_corrupt + 1) % conf->gen_report_threshold == 0) {
				
				MTIMER_STOP(process_timer);
				MTIMER_CALC(process_timer);
				
				printf("]\n");
				
				MTIMER_START(post_process_timer);
				M_DEBUG2(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_WARNINGS,
					"o(t) writing month %02i - %04i\n", l_month+1, l_year+1900);
				
				l = state_list;
				
				while (l) {
					mdata *data = l->data;
					mdata *histdata = NULL;
					
					if (!data) break;
					
					histdata = mdata_WebHist_create_by_state(data->data.state.state);
				
					mlist_insert_replace(data->data.state.history, histdata);
					
					mstate_write(conf, data->data.state.state, M_STATE_WRITE_BY_MONTH, *(data->key) ? data->key : NULL);
					history_write(conf, data->data.state.history, *(data->key) ? data->key : NULL);
					
					if (0 != generate_monthly_output(conf, data->data.state.state, *(data->key) ? data->key : NULL)) {
						fprintf(stderr, "%s.%d: output generation failed - no data this month?\n", __FILE__, __LINE__);
						//return -1;
					}
					if (0 != generate_history_output(conf, data->data.state.history, *(data->key) ? data->key : NULL)) {
						fprintf(stderr, "%s.%d: history generation failed - going down hard\n", __FILE__, __LINE__);
						return -1;
					}
	
					l = l->next;
				}
				MTIMER_STOP(post_process_timer);
				MTIMER_CALC(post_process_timer);
				
				
				printf("[");
				MTIMER_START(process_timer);
			}

			if (tm->tm_mon != l_month || tm->tm_year != l_year) {
				
				MTIMER_STOP(process_timer);
				MTIMER_CALC(process_timer);
				
				printf("]\n");
				
				MTIMER_START(post_process_timer);
				
				M_DEBUG2(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_WARNINGS,
					"o writing month %02i - %04i\n", l_month+1, l_year+1900);
				
				
				l = state_list;
				
				while (l) {
					mdata *data = l->data;
					mdata *histdata = NULL;
					
					if (!data) break;
					
					histdata = mdata_WebHist_create_by_state(data->data.state.state);
				
					mlist_insert_replace(data->data.state.history, histdata);
					
					if (conf->debug_level > 3)
						fprintf(stderr, "o(t) -- state written: (...)/%s/\n", *(data->key) ? data->key : ".");
					mstate_write(conf, data->data.state.state, M_STATE_WRITE_BY_MONTH, *(data->key) ? data->key : NULL);
					
					if (0 != generate_monthly_output(conf, data->data.state.state, *(data->key) ? data->key : NULL)) {
						fprintf(stderr, "%s.%d: output generation failed - no data this month?\n", __FILE__, __LINE__);
						//return -1;
					}
					
					mstate_free(data->data.state.state);
					data->data.state.state = mstate_init();
					
					l = l->next;
				}
				
				tm = localtime(&(rec->timestamp));
			
				l_month = tm->tm_mon;
				l_year = tm->tm_year;
				
				MTIMER_STOP(post_process_timer);
				MTIMER_CALC(post_process_timer);
				
				for (i = 0; i < ((lines % 50000) / 1000); i++) printf(" ");
				printf("[");
				
				MTIMER_START(process_timer);
			}

			insert_record(conf, state_list, rec);
			
			first_valid_line = 1;
			
			l_stamp = rec->timestamp;
		} else if (ret == M_RECORD_SKIPPED) {
			lines_skipped++;
		} else if (ret == M_RECORD_IGNORED) {
			lines_ignored++;
		} else if (ret == M_RECORD_HARD_ERROR) {
			M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
				 " a hard error occured in line %ld - going down hard\n", lines);
			
			exit(-1);
		} else {
			lines_corrupt++;
			
			M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
				 " parser reported an error in line %ld\n", lines);
		}
		mrecord_free(rec);
		
		
		/* cosmetics */
		if (lines % 1000 == 0) {
			printf("."); 
			if (lines % (1000 * 50)== 0) {
				printf(" %8ld", lines);
				if (conf->debug_level > 1) {
					printf(" - %10.2f (%4ld, %4ld)",
					       (lines)/((MTIMER_GET_MSEC(parse_timer)+MTIMER_GET_MSEC(process_timer))/1000.0), 
					       lines_corrupt,
					       lines_skipped);
				}
				printf("\n ");
			}
			fflush(stdout); 
		}
		
		rec = mrecord_init();
		
		MTIMER_STOP(process_timer);
		MTIMER_CALC(process_timer);
		
		MTIMER_START(parse_timer);
	}

	MTIMER_STOP(parse_timer);
	MTIMER_CALC(parse_timer);

	printf("]\n ");

	MTIMER_START(post_process_timer);

	if (l_month != -1) {
		int create_bp = 1;
		FILE *split_file;
		char *fn;

		if (conf->debug_level > 0) 
			printf("writing month %02i - %04i\n",l_month+1, l_year+1900);
		
		if (conf->debug_level > 1)
			fprintf(stderr, "%s.%d: writing the last month\n", __FILE__, __LINE__ );
		
		/* create filename */
		fn = malloc(strlen(conf->outputdir) + strlen("/modlogan.statefiles") + 1);
		strcpy(fn, conf->outputdir);
		strcat(fn, "/modlogan.statefiles");
		split_file = fopen(fn, "w");
			
		l = state_list;
				
		while (l) {
			mdata *data = l->data;
			mdata *histdata = NULL;
			
			if (!data) break;
			
			if (conf->debug_level > 1)
				fprintf(stderr, "%s.%d: subpath = %s\n", __FILE__, __LINE__, data->key);
			
			histdata = mdata_WebHist_create_by_state(data->data.state.state);
				
			mlist_insert_replace(data->data.state.history, histdata);
			
			mstate_write(conf, data->data.state.state, M_STATE_WRITE_DEFAULT, *(data->key) ? data->key : NULL);
			history_write(conf, data->data.state.history, *(data->key) ? data->key : NULL);

			if (0 != generate_monthly_output(conf, data->data.state.state, *(data->key) ? data->key : NULL)) {
				fprintf(stderr, "%s.%d: output generation failed - no data this month?\n", __FILE__, __LINE__);
				//return -1;
			}
			if (0 != generate_history_output(conf, data->data.state.history, *(data->key) ? data->key : NULL)) {
				fprintf(stderr, "%s.%d: history generation failed - going down hard\n", __FILE__, __LINE__);
				return -1;
			}

			l = l->next;
			
			/* disable the bigpicture if we aren't splitting the logs */
			if (*(data->key) == '\0') create_bp = 0;
			
			/* write the current directoy to the split file */
			fprintf(split_file, "%s\n", data->key);
			
		}
		
		fclose(split_file);
		free(fn);
		
		/* combine the statefiles and generate the BIG picture */
		if (create_bp) {
			/* BIG picture */
		}
	}
	
	MTIMER_STOP(post_process_timer);
	MTIMER_CALC(post_process_timer);

	mlist_free(state_list);
	mrecord_free(rec);
	
	if (conf->debug_level > 0) {
		printf(" --> Setup       : %10.2fs <--\n", MTIMER_GET_MSEC(setup_timer)/1000.0);
		printf(" --> Parse       : %10.2fs <--\n", MTIMER_GET_MSEC(parse_timer)/1000.0);
		printf(" --> Process     : %10.2fs <--\n", MTIMER_GET_MSEC(process_timer)/1000.0);
		printf(" --> Post-Process: %10.2fs <--\n", MTIMER_GET_MSEC(post_process_timer)/1000.0);
		
		printf("%s: %.2f %s (%ld %s, %ld %s, %ld %s, %ld %s)\n",
		       "Throughput", (lines)/((MTIMER_GET_MSEC(parse_timer)+MTIMER_GET_MSEC(process_timer))/1000.0), 
		       "rec/s",
		       lines, "records", 
		       lines_corrupt, "corrupt records",
		       lines_skipped, "skipped records",
		       lines_ignored, "ignored records"
		       );
	}
	
	mconfig_free(conf);
	return 0;
}
