/*
** Modular Logfile Analyzer
** Copyright 2000 Jan Kneschke <jan@kneschke.de>
**
** Homepage: http://www.kneschke.de/projekte/modlogan
**

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version, and provided that the above
    copyright and permission notice is included with all distributed
    copies of this or derived software.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

**
** $Id: mdatatypes.c,v 1.39 2001/09/15 19:06:44 ostborn Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <libxml/tree.h>

#include "config.h"
#include "mdatatypes.h"
#include "misc.h"

/* the datatypes */
#include "datatypes/count/datatype.h"
#include "datatypes/visited/datatype.h"
#include "datatypes/webhist/datatype.h"
#include "datatypes/state/datatype.h"
#include "datatypes/match/datatype.h"
#include "datatypes/record/datatype.h"
#include "datatypes/sublist/datatype.h"
#include "datatypes/split/datatype.h"
#include "datatypes/brokenlink/datatype.h"
#include "datatypes/traffic/datatype.h"
#include "datatypes/netmask/datatype.h"
#include "datatypes/visit/datatype.h"
#include "datatypes/query/datatype.h"

typedef struct {
	int type;
	const char *string;
} key_map;

const key_map key_mapping[] = {
	{ M_DATA_TYPE_COUNT, "count" },
	{ M_DATA_TYPE_VISITED, "visited" },
	{ M_DATA_TYPE_WEBHIST, "webhist" },
	{ M_DATA_TYPE_BROKENLINK, "brokenlink" },
	{ M_DATA_TYPE_VISIT, "visit" },
	{ M_DATA_TYPE_SUBLIST, "sublist" },
	
	{ -1, NULL }
};

/**
 * get a string represenation for the current datatype
 * 
 * mainly used by the xml-write code (key='...')
 * 
 * creates warnings if the datatype is unknown, not handled or unset
 * 
 * @param data the datatype 
 * @return the corresponding string or 'unknown' if the datatype is currently not handled
 */
const char *mdata_get_key_from_datatype(mdata *data) {
	int i;
	if (!data) return NULL;
	
	i = 0;
	while (key_mapping[i].string != NULL) {
		if (key_mapping[i].type == data->type) {
			return key_mapping[i].string;
		}
		i++;
	}
	
	/* if everything went well we shouldn't reach this point */
	fprintf(stderr, "%s.%d: can't %s datatype '%d' (%s)\n", __FILE__, __LINE__, "get key for", data->type, data->key);

	return "unknown";
}

/**
 * get the data for a given string representation
 * 
 * mainly used by the xml-read code (key='...')
 * 
 * creates warnings if the datatype is unknown, not handled or unset
 * 
 * @param data the datatype 
 * @return the corresponding string or 'unknown' if the datatype is currently not handled
 */
int mdata_get_datatype_from_key(const char *key) {
	int i;
	if (!key) return M_DATA_TYPE_UNSET;
	
	i = 0;
	while (key_mapping[i].string != NULL) {
		if (strcmp(key_mapping[i].string, key) == 0) {
			return key_mapping[i].type;
		}
		i++;
	}
	
	/* if everything went well we shouldn't reach this point */
	fprintf(stderr, "%s.%d: can't %s datatype '%s'\n", __FILE__, __LINE__, "get datatype for", key);

	return M_DATA_TYPE_UNSET;
}

/**
 * free all structures of the datatype
 * 
 * creates warnings if the datatype is unknown, not handled or unset
 * 
 * @param data the to be freed datatype
 */
void mdata_free(mdata *data) {
	if (!data) return;
	
	switch(data->type) {
		case M_DATA_TYPE_COUNT: 
			mdata_Count_free(data);
			break;
		case M_DATA_TYPE_VISITED:
			mdata_Visited_free(data);
			break;
		case M_DATA_TYPE_WEBHIST:
			mdata_WebHist_free(data);
			break;
		case M_DATA_TYPE_RECORD:
			mdata_Record_free(data);
			break;
		case M_DATA_TYPE_STATE:
			mdata_State_free(data);
			break;
		case M_DATA_TYPE_MATCH:
			mdata_Match_free(data);
			break;
		case M_DATA_TYPE_SUBLIST:
			mdata_SubList_free(data);
			break;
		case M_DATA_TYPE_SPLIT:
			mdata_Split_free(data);
			break;
		case M_DATA_TYPE_BROKENLINK:
			mdata_BrokenLink_free(data);
			break;
		case M_DATA_TYPE_TRAFFIC:
			mdata_Traffic_free(data);
			break;
		case M_DATA_TYPE_NETMASK:
			mdata_Netmask_free(data);
			break;
		case M_DATA_TYPE_VISIT:
			mdata_Visit_free(data);
			break;
		case M_DATA_TYPE_QUERY:
			mdata_Query_free(data);
			break;
		case M_DATA_TYPE_UNSET:
			fprintf(stderr, "%s.%d: can't %s unset datatype\n", __FILE__, __LINE__, "free");
			break;
		default:
			fprintf(stderr, "%s.%d: can't %s datatype '%d'\n", __FILE__, __LINE__, "free", data->type);
			break;
	}
	
	if (data->key)	free(data->key);
#ifdef DEBUG_DATATYPES
	fprintf(stderr, "%s.%d: freeing ID '%d'\n", __FILE__, __LINE__, data->id);
#endif
	free(data);
}

/**
 * create a new datatype
 * 
 * allocates the memory for a new datatype and sets the internal 
 * type specifier to UNSET
 * 
 * @return a new 'undefined' datatype
 */

mdata *mdata_init () {
	mdata *data;
	static int id = 0;
	data = malloc(sizeof(mdata));
	assert(data);
	
	data->key	= NULL;
	data->type	= M_DATA_TYPE_UNSET;
	data->id        = id++;
	
	return data;
}

/**
 * convertes the datatype to XML
 * 
 * this function is mainly a dispatcher between the different datatypes
 * and just calles their own ..._to_xml functions.
 * 
 * creates warnings if the datatype is unknown, not handled or unset
 * 
 * @param node a XML node
 * @param data the datatype which 
 * @return -1 on error or the return value of the specific _to_xml function
 */

int mdata_datatype_to_xml(gzFile *fd, mdata *data) {
	if (!data) return -1;
	
	switch (data->type) {
		case M_DATA_TYPE_COUNT: 
			return mdata_Count_to_xml(fd, data);
		case M_DATA_TYPE_VISITED: 
			return mdata_Visited_to_xml(fd, data);
		case M_DATA_TYPE_BROKENLINK: 
			return mdata_BrokenLink_to_xml(fd, data);
		case M_DATA_TYPE_WEBHIST: 
			return mdata_WebHist_to_xml(fd, data);
		case M_DATA_TYPE_VISIT: 
			return mdata_Visit_to_xml(fd, data);
		case M_DATA_TYPE_SUBLIST: 
			return mdata_SubList_to_xml(fd, data);
		case M_DATA_TYPE_UNSET:
			fprintf(stderr, "%s.%d: can't %s unset datatype\n", __FILE__, __LINE__, "write");
			break;
		default:
			fprintf(stderr, "%s.%d: can't %s datatype '%d'\n", __FILE__, __LINE__, "write", data->type);
			break;
	}
	return -1;
}

/**
 * creates a new XML-node in the XML-tree and inserts the XML from the datatypes
 * 
 * @param node a XML node
 * @param data the datatype which
 */

int mdata_write(gzFile *fd, mdata *data) {
	// encode key with url-encoding
	char *encoded_key = url_encode(data->key);
	
	gzprintf(fd, "<%s key=\"%s\">\n", mdata_get_key_from_datatype(data), encoded_key);
	// free memory
	free(encoded_key);
	
	mdata_datatype_to_xml(fd, data);
	
	gzprintf(fd, "</%s>\n", mdata_get_key_from_datatype(data));
	
	return 0;
}

/**
 * combine two elements 
 * 
 * used if the keys of both elements are equal while a *_insert 
 * the actual append function can reject the combine if the internal contrains
 * aren't matched
 * 
 * @param dst elements 
 * @param src element which 
 * @return M_DATA_APPEND - append successfull
 */

mdata_append_return mdata_append(mdata *dst, mdata *src) {
	int ret = 0;
	if (dst->type != src->type) return -1;
	
	switch(src->type) {
		case M_DATA_TYPE_COUNT: 
			ret = mdata_Count_append(dst, src);
			break;
		case M_DATA_TYPE_VISITED:
			ret = mdata_Visited_append(dst, src);
			break;
		case M_DATA_TYPE_VISIT:
			ret = mdata_Visit_append(dst, src);
			break;
		case M_DATA_TYPE_WEBHIST:
			ret = mdata_WebHist_append(dst, src);
			break;
		case M_DATA_TYPE_RECORD:
			ret = mdata_Record_append(dst, src);
			break;
		case M_DATA_TYPE_STATE:
			ret = mdata_State_append(dst, src);
			break;
		case M_DATA_TYPE_MATCH:
			ret = mdata_Match_append(dst, src);
			break;
		case M_DATA_TYPE_SUBLIST:
			ret = mdata_SubList_append(dst, src);
			break;
		case M_DATA_TYPE_SPLIT:
			ret = mdata_Split_append(dst, src);
			break;
		case M_DATA_TYPE_BROKENLINK:
			ret = mdata_BrokenLink_append(dst, src);
			break;
		case M_DATA_TYPE_TRAFFIC:
			ret = mdata_Traffic_append(dst, src);
			break;
		case M_DATA_TYPE_NETMASK:
			ret = mdata_Netmask_append(dst, src);
			break;
		case M_DATA_TYPE_UNSET:
			fprintf(stderr, "%s.%d: can't %s unset datatype\n", __FILE__, __LINE__, "append");
			break;
		default:
			fprintf(stderr, "%s.%d: can't %s datatype '%d'\n", __FILE__, __LINE__, "append", src->type);
			break;
	}
	
	return ret;
}

/**
 * copy the element
 * 
 * @param src to be copied element
 * @return a copy of the provided element
 */

mdata *mdata_copy(mdata *src) {
	mdata *dst = NULL;
	
	switch(src->type) {
		case M_DATA_TYPE_COUNT: 
			dst = mdata_Count_copy(src);
			break;
		case M_DATA_TYPE_VISITED:
			dst = mdata_Visited_copy(src);
			break;
		case M_DATA_TYPE_WEBHIST:
			dst = mdata_WebHist_copy(src);
			break;
		case M_DATA_TYPE_RECORD:
			dst = mdata_Record_copy(src);
			break;
		case M_DATA_TYPE_STATE:
			dst = mdata_State_copy(src);
			break;
		case M_DATA_TYPE_MATCH:
			dst = mdata_Match_copy(src);
			break;
		case M_DATA_TYPE_SUBLIST:
			dst = mdata_SubList_copy(src);
			break;
		case M_DATA_TYPE_SPLIT:
			dst = mdata_Split_copy(src);
			break;
		case M_DATA_TYPE_BROKENLINK:
			dst = mdata_BrokenLink_copy(src);
			break;
		case M_DATA_TYPE_TRAFFIC:
			dst = mdata_Traffic_copy(src);
			break;
		case M_DATA_TYPE_NETMASK:
			dst = mdata_Netmask_copy(src);
			break;
		case M_DATA_TYPE_UNSET:
			fprintf(stderr, "%s.%d: can't %s unset datatype\n", __FILE__, __LINE__, "copy");
			break;
		default:
			fprintf(stderr, "%s.%d: can't %s datatype '%d'\n", __FILE__, __LINE__, "copy", src->type);
			break;
	} 
	
	return dst;
}

/**
 * convert a xml-node to a 'mdata'
 * 
 * spites error messages if a unhandled datatype should be read
 * 
 * return 0 - success, -1 error
 * 
 */
int mdata_read(mdata *data, xmlNodePtr node) {
	char *encoded_key;
	char *decoded_key;
	if (node == NULL) return -1;
//	fprintf(stderr, "%s.%d: hittin %s\n", __FILE__, __LINE__, node->name);
	if (node->type != XML_ELEMENT_NODE) return -1;

	// uri-decode the key
	encoded_key = xmlGetProp(node, "key");
	decoded_key = url_decode(encoded_key);
	// we don't need the encoded anymore
	free(encoded_key);
	
	switch(mdata_get_datatype_from_key(node->name)) {
	case M_DATA_TYPE_COUNT:
		mdata_Count_from_xml(data, node->children, decoded_key);
		break;
	case M_DATA_TYPE_VISITED:
		mdata_Visited_from_xml(data, node->children, decoded_key);
		break;
	case M_DATA_TYPE_BROKENLINK:
		mdata_BrokenLink_from_xml(data, node->children, decoded_key);
		break;
	case M_DATA_TYPE_VISIT:
		mdata_Visit_from_xml(data, node->children, decoded_key);
		break;
	case M_DATA_TYPE_WEBHIST:
		mdata_WebHist_from_xml(data, node->children, decoded_key);
		break;
	case M_DATA_TYPE_SUBLIST:
		mdata_SubList_from_xml(data, node->children, decoded_key);
		break;
	case M_DATA_TYPE_UNSET:
		fprintf(stderr, "%s.%d: can't %s unset datatype\n", __FILE__, __LINE__, "read");
		break;
	default:
		fprintf(stderr, "%s.%d: can't %s xml for datatype '%s'\n", __FILE__, __LINE__, "read", node->name);
	}
	
	// free the memory of the decoded key
	free(decoded_key);
	return 0;
}

/**
 * get the count of the countable elements
 * 
 * @param data the element which is queried
 * @return the count value or 0 if the uncountable value is queried
 */

int mdata_get_count(mdata *data) {
	if (data == NULL) return 0;

	switch(data->type) {
		case M_DATA_TYPE_COUNT:
			return data->data.count.count;
		case M_DATA_TYPE_VISITED:
			return data->data.visited.count;
		case M_DATA_TYPE_BROKENLINK:
			return data->data.brokenlink.count;
		case M_DATA_TYPE_SUBLIST:
			return data->data.sublist.count;
		case M_DATA_TYPE_UNSET:
			fprintf(stderr, "%s.%d: can't %s from unset datatype (%s)\n", __FILE__, __LINE__, "get count", data->key);
			break;
		default:
			fprintf(stderr, "%s.%d: can't %s from datatype '%d'\n", __FILE__, __LINE__, "get count", data->type);
	}
	return 0;
}

/**
 * set the count of countable elements
 * 
 * @param data
 * @param count
 * @return -1 error, 0 ok
 */

int mdata_set_count(mdata *data, int count) {
	if (data == NULL) return -1;

	switch(data->type) {
		case M_DATA_TYPE_COUNT:
			data->data.count.count = count;
			return 0;
		case M_DATA_TYPE_VISITED:
			data->data.visited.count = count;
			return 0;
		case M_DATA_TYPE_BROKENLINK:
			data->data.brokenlink.count = count;
			return 0;
		case M_DATA_TYPE_SUBLIST:
			data->data.sublist.count = count;
			return 0;
		case M_DATA_TYPE_UNSET:
			fprintf(stderr, "%s.%d: can't %s for unset datatype\n", __FILE__, __LINE__, "set count");
			break;
		default:
			fprintf(stderr, "%s.%d: can't %s for datatype '%d'\n", __FILE__, __LINE__, "set count", data->type);
	}
	return -1;
}

/**
 * transform a xmlleaf to the internals strutures
 * 
 * @param dest destination
 * @param type type of the destination
 * @param node the xml leaf
 * @return
 */

int mdata_insert_value(void *dest, mdata_field_type type, xmlNodePtr node) {
	switch (type) {
	case M_DATA_FIELDTYPE_LONG: {
		int str = *(int *)(dest);
		char *buffer;
			
		if ((buffer = xmlNodeGetContent(node)) != NULL) {
			
			str = strtol(buffer, NULL, 10);
			
			*(int *)(dest) = str;
			
//			fprintf(stderr, "%s.%d: xml-read: %d\n", __FILE__, __LINE__, str);
			
			free(buffer);
		} else {
			fprintf(stderr, "%s.%d: xml-read failed (long) -> %s\n", __FILE__, __LINE__, node != NULL ? node->name : "(NULL)");
		}
			
		break;
	}
	case M_DATA_FIELDTYPE_DOUBLE: {
		double str = *(double *)(dest);
		char *buffer;
			
		if ((buffer = xmlNodeGetContent(node)) != NULL) {
			
			str = strtod(buffer, NULL);
			
			*(double *)(dest) = str;
			
//			fprintf(stderr, "%s.%d: xml-read: %d\n", __FILE__, __LINE__, str);
			
			free(buffer);
		} else {
			fprintf(stderr, "%s.%d: xml-read failed (double) -> %s\n", __FILE__, __LINE__, node ? node->name : "(NULL)");
		}
			
		break;
	}
	case M_DATA_FIELDTYPE_STRING: {
		char *buffer;
			
		if ((buffer = xmlNodeGetContent(node)) != NULL) {
			char *str;
			
			if (*(char **)(dest) != NULL) {
				free(*(char **)(dest));
				*(char **)(dest) = NULL;
			}
			
			str = malloc(strlen(buffer) + 1);
			strcpy(str, buffer);
			
			*(char **)(dest) = str;
			
//			fprintf(stderr, "%s.%d: xml-read: %s\n", __FILE__, __LINE__, str);
			
			free(buffer);
		//} else {
			//fprintf(stderr, "%s.%d: xml-read failed (string) -> %s\n", __FILE__, __LINE__, node ? node->name : "(NULL)");
		}
			
		break;
	}
	
	case M_DATA_FIELDTYPE_HASH: {
		mdata *data;
		mhash *hash = *(mhash **) dest;
		
		while (node) {
			if (node->type != XML_ELEMENT_NODE) {
				node = node->next;
				continue;
			}
			
			data = mdata_init();
#ifdef DEBUG_DATATYPES
			fprintf(stderr, "%s.%d: got %d\n", __FILE__, __LINE__, data->id);
#endif
			mdata_read(data, node);
		
			if (data->key) {
				mhash_insert(hash, data);
			} else {
				fprintf(stderr, "%s.%d: *args* no key - going down hard\n", __FILE__, __LINE__);
				return -1;
			}
			
			node = node->next;
		}
		
		break;
	}
		
	case M_DATA_FIELDTYPE_LIST: {
		mdata *data;
		mlist *list = *(mlist **) dest;
		
		while (node) {
			if (node->type != XML_ELEMENT_NODE) {
				node = node->next;
				continue;
			}
			
			data = mdata_init();
#ifdef DEBUG_DATATYPES
			fprintf(stderr, "%s.%d: got %d\n", __FILE__, __LINE__, data->id);
#endif
			mdata_read(data, node);
		
			if (data->key) {
				mlist_insert(list, data);
			} else {
				fprintf(stderr, "%s.%d: *args* no key - going down hard\n", __FILE__, __LINE__);
				return -1;
			}
			
			node = node->next;
		}
		
		break;
	}

	default: 
		fprintf(stderr, "%s.%d: insert failed: unknown type %d -> %s\n", 
			__FILE__, __LINE__, 
			type,
			node ? node->name : "(NULL)");
		return -1;
	}
	return 0;
}
