/*
** Modular Logfile Analyzer
** Copyright 2000 Jan Kneschke <jan@kneschke.de>
**
** Homepage: http://www.kneschke.de/projekte/modlogan
**

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version, and provided that the above
    copyright and permission notice is included with all distributed
    copies of this or derived software.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

**
** $Id: mhash.c,v 1.10 2001/05/19 23:27:30 jk Exp $
*/

#include <libintl.h>
#include <locale.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "mlist.h"
#include "mhash.h"
#include "mdatatypes.h"

int mhash_unfold(mhash *h, mlist *l ) {
	int i = 0;
	for ( i = 0; i < h->size; i++) {
		if (h->data[i]->list) {
			mlist *hl;
			
			hl = h->data[i]->list;
			while (hl) {
				if (hl->data) {
					mdata *data = mdata_copy(hl->data);
				
					mlist_insert(l, data);
				}
				hl = hl->next;
			}
		}
	}
	return 0;
}

int mhash_unfold_sorted_limited(mhash *h, mlist *l, int count ) {
	int i, j, max;
	mdata *data = NULL, *ins_data;
 
	for ( j = 0; j < count; j++) {
		data = NULL;
 
		max = 0;
 
		for ( i = 0; i < h->size; i++) {
			if (h->data[i]->list) {
				mlist *hl;

				hl = h->data[i]->list;
				while (hl) {
					if (hl->data) {
						if (mdata_get_count(hl->data) > max) {
							max = mdata_get_count(hl->data);
							data = hl->data;
						}
					}
					hl = hl->next;
				}
			}
		}

		if (data) {
			ins_data = mdata_copy(data);
 
			mlist_insert(l, ins_data);
	/* marking element as processed */
			mdata_set_count(data, - mdata_get_count(data));
		}
	}

	/* resetting count */

	for ( i = 0; i < h->size; i++) {
		if (h->data[i]->list) {
			mlist *hl;
 
			hl = h->data[i]->list;
			while (hl) {
				if (hl->data) {
					if (mdata_get_count(hl->data) < 0) {
						data = hl->data;
						
						mdata_set_count(hl->data, - mdata_get_count(hl->data));
					}
				}
				hl = hl->next;
			}
		}
	}
 
	return 0;
}

int mhash_calc (mhash *h, const char *str) {
	unsigned long i = 0;
	
	while (*str) 
		i = *str++ + 31 * i;
	
	return (i % h->size);
}

mhash *mhash_init (int size) {
	int i = 0;
	
	mhash *h;
	
	h = malloc(sizeof(mhash));
	assert(h);
	
	h->size = size;
	h->data = malloc(sizeof(mhash_data *) * h->size);
	
	for (i = 0; i < h->size; i++) {
		h->data[i] = malloc(sizeof(mhash_data));
		assert(h->data[i]);
		h->data[i]->list = mlist_init();
	}
	
	return h;
}

int mhash_free (mhash *h) {
	int i = 0;
	
	for (i = 0; i < h->size; i++) {
		mlist_free(h->data[i]->list);
		free(h->data[i]);
	}
	free(h->data);
	free(h);
	
	return 0;
}

int mhash_insert(mhash *h, mdata *data) {
	int ndx = 0;

#define MHASH_DEPTH_THRESHOLD 32
//#define DEBUG_MHASH
#ifdef DEBUG_MHASH
	int max_cnt = 0;
	int i;
	int cnt;
	
	for ( i = 0; i < h->size; i++) {
		cnt = mlist_count(h->data[i]->list);
		if (cnt >= MHASH_DEPTH_THRESHOLD && 
			cnt > max_cnt) {
			max_cnt = cnt;
		}
	}
	
	if (max_cnt) {
		fprintf(stderr, "%s.%d: hash inc ?? |%d| %d > %d (%p)\n", 
			__FILE__, __LINE__, 
			h->size, max_cnt, MHASH_DEPTH_THRESHOLD,
			h);
	}
	
#endif
	
	if (data->key == NULL) return -1;
	
	ndx = mhash_calc(h, data->key);
	
	mlist_insert(h->data[ndx]->list, data);
	
	return 0;
}

int mhash_insert_sorted(mhash *h, mdata *data) {
	int ndx = 0;

#define MHASH_DEPTH_THRESHOLD 32
//#define DEBUG_MHASH
#ifdef DEBUG_MHASH
	int max_cnt = 0;
	int i;
	int cnt;
	
	for ( i = 0; i < h->size; i++) {
		cnt = mlist_count(h->data[i]->list);
		if (cnt >= MHASH_DEPTH_THRESHOLD && 
			cnt > max_cnt) {
			max_cnt = cnt;
		}
	}
	
	if (max_cnt) {
		fprintf(stderr, "%s.%d: hash inc ?? |%d| %d > %d (%p)\n", 
			__FILE__, __LINE__, 
			h->size, max_cnt, MHASH_DEPTH_THRESHOLD,
			h);
	}
	
#endif
	
	if (data->key == NULL) return -1;
	
	ndx = mhash_calc(h, data->key);
	
	mlist_insert_sorted(h->data[ndx]->list, data);
	
	return 0;
}


int mhash_count(mhash *h) {
	int i, c = 0;
	if (!h) return 0;
	
	for ( i = 0; i < h->size; i++) {
		c += mlist_count(h->data[i]->list);
	}
	
	return c;
}

mdata *mhash_get_data(mhash *h, const char *str) {
	int ndx;
	
	ndx = mhash_calc(h, str);
	
	return mlist_get_data(h->data[ndx]->list, str);
}

int mhash_in_hash(mhash *h, const char *str) {
	int ndx;
	
	ndx = mhash_calc(h, str);
	
	return mlist_in_list(h->data[ndx]->list, str);
}
