/* 
** Modular Logfile Analyzer
** Copyright 2000 Jan Kneschke <jan@kneschke.de>
**
** Homepage: http://www.kneschke.de/projekte/modlogan
**

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version, and provided that the above
    copyright and permission notice is included with all distributed
    copies of this or derived software.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

**
** $Id: mstate.c,v 1.36 2001/09/09 16:00:22 lohgan Exp $
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>

#include <libxml/tree.h>

#include "config.h"
#include "mlist.h"
#include "mstate.h"
#include "mdatatypes.h"
#include "mlocale.h"

#define STATE_FILE_VERSION	"0.5"

mstate *mstate_init() {
	mstate *state = malloc(sizeof(mstate));
	assert(state);
	
	state->year		= 0;
	state->month		= 0;
	state->timestamp	= 0;
	
	state->ext		= NULL;
	state->ext_type		= M_STATE_TYPE_UNSET;
	
	return state;
}

void mstate_free(mstate *state) {
	if (!state) return;
	
	if (state->ext) {
		switch(state->ext_type) {
		case M_STATE_TYPE_WEB:
			mstate_free_web(state->ext);
			break;
		case M_STATE_TYPE_TELECOM:
			mstate_free_telecom(state->ext);
			break;
		case M_STATE_TYPE_TRAFFIC:
			mstate_free_traffic(state->ext);
			break;
		case M_STATE_TYPE_MAIL:
			mstate_free_mail(state->ext);
			break;
		case M_STATE_TYPE_UNSET:
			break;
		default:
			fprintf(stderr, "%s.%d: unknown substate type %d\n", __FILE__, __LINE__, state->ext_type);
		}
	}
	
	free(state);
}

mstate_web *mstate_init_web() {
	mstate_web *state = malloc(sizeof(mstate_web));
	assert(state);
	memset(state, 0, sizeof(mstate_web));
	
	state->visit_list	= mlist_init();
	state->indexed_pages	= mhash_init( 32 );
	state->os_hash		= mhash_init( 32 );
	state->ua_hash		= mhash_init( 32 );
	state->req_url_hash	= mhash_init( 128 );
	state->req_prot_hash	= mhash_init( 16 );
	state->req_meth_hash	= mhash_init( 16 );
	state->status_hash	= mhash_init( 16 );
	state->host_hash	= mhash_init( 256 );
	state->ref_url_hash	= mhash_init( 64 );
	state->robots		= mhash_init( 32 );
	state->bookmarks	= mhash_init( 32 );
	state->status_internal_error = mhash_init( 32 );
	state->status_missing_file = mhash_init( 32 );
	state->searchstring	= mhash_init( 32 );
	state->searchsite	= mhash_init( 32 );
	state->country_hash	= mhash_init( 32 );
	state->extension	= mhash_init( 32 );
	state->visits     	= mhash_init( 128 );
	state->views     	= mhash_init( 128 );
	state->vhost_hash	= mhash_init( 32 );

	return state;
}

void mstate_free_web(mstate_web *state) {
	if (!state) return;
	
	mhash_free(state->os_hash);
	mhash_free(state->ua_hash);
	mhash_free(state->req_prot_hash);
	mhash_free(state->req_meth_hash);
	mhash_free(state->status_hash);
	mhash_free(state->host_hash);
	mhash_free(state->req_url_hash);
	mhash_free(state->ref_url_hash);
	mhash_free(state->bookmarks);
	mhash_free(state->robots);
	mhash_free(state->status_internal_error);
	mhash_free(state->status_missing_file);
	mhash_free(state->searchstring);
	mhash_free(state->searchsite);
	mhash_free(state->country_hash);
	mhash_free(state->indexed_pages);
	mhash_free(state->extension);
	mhash_free(state->visits);
	mhash_free(state->views);
	mhash_free(state->vhost_hash);
	
	mlist_free(state->visit_list);
	
	free(state);
}

mstate_telecom *mstate_init_telecom() {
	int i;
	
	mstate_telecom *state = malloc(sizeof(mstate_telecom));
	assert(state);
	
	state->called_numbers	= mhash_init( 128 );
	state->calling_numbers	= mhash_init( 128 );
	
	for (i = 0; i < 24; i++) {
		state->hours[i].incomming_calls = 0;
		state->hours[i].outgoing_calls = 0;
	}
	
	for (i = 0; i < 31; i++) {
		state->days[i].incomming_calls = 0;
		state->days[i].outgoing_calls = 0;
	}
	
	return state;
}

void mstate_free_telecom(mstate_telecom *state) {
	if (!state) return;
	
	mhash_free(state->called_numbers);
	mhash_free(state->calling_numbers);
	
	free(state);
}

mstate_mail *mstate_init_mail() {
	mstate_mail *state = malloc(sizeof(mstate_mail));
	assert(state);
	
	memset(state, 0, sizeof(mstate_mail));
	
	state->sender = mhash_init( 128 );
	state->receipient = mhash_init( 128 );
	
	return state;
}

void mstate_free_mail(mstate_mail *state) {
	if (!state) return;
	
	mhash_free(state->sender);
	mhash_free(state->receipient);
	
	free(state);
}

mstate_traffic *mstate_init_traffic() {
	mstate_traffic *state = malloc(sizeof(mstate_traffic));
	assert(state);
	
	state->incomming	= mhash_init( 128 );
	state->outgoing		= mhash_init( 128 );
	state->external		= mhash_init( 128 );
	state->internal		= mhash_init( 128 );
	
	return state;
}

void mstate_free_traffic(mstate_traffic *state) {
	if (!state) return;
	
	mhash_free(state->incomming);
	mhash_free(state->outgoing);
	mhash_free(state->external);
	mhash_free(state->internal);
	
	free(state);
}

int mhash_write(gzFile *fd, mhash *h) {
	int i;

	for (i = 0; i < h->size; i++) {
		if (h->data[i]->list) {
			mlist_write(fd, h->data[i]->list);
		}
	}
	
	return 0;
}

int mstate_write_web (mconfig *conf, mstate_web *state, gzFile *fd) {
	int i;
	
	gzprintf(fd, "<%s>\n", M_STATE_SECTION_WEB);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_REQ_URL);
	mhash_write(fd, state->req_url_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_REQ_URL);

	gzprintf(fd, "<%s>\n", M_STATE_WEB_REQ_METHOD);
	mhash_write(fd, state->req_meth_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_REQ_METHOD);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_REQ_PROTOCOL);
	mhash_write(fd, state->req_prot_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_REQ_PROTOCOL);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_REF_URL);
	mhash_write(fd, state->ref_url_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_REF_URL);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_OS);
	mhash_write(fd, state->os_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_OS);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_USERAGENT);
	mhash_write(fd, state->ua_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_USERAGENT);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_HOST);
	mhash_write(fd, state->host_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_HOST);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_STATUS);
	mhash_write(fd, state->status_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_STATUS);

	gzprintf(fd, "<%s>\n", M_STATE_WEB_VISIT);
	mlist_write(fd, state->visit_list);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_VISIT);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_COUNTRIES);
	mhash_write(fd, state->country_hash);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_COUNTRIES);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_ROBOTS);
	mhash_write(fd, state->robots);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_ROBOTS);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_SEARCHSITE);
	mhash_write(fd, state->searchsite);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_SEARCHSITE);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_SEARCHSTRING);
	mhash_write(fd, state->searchstring);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_SEARCHSTRING);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_INT_ERROR);
	mhash_write(fd, state->status_internal_error);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_INT_ERROR);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_MISSING_FILE);
	mhash_write(fd, state->status_missing_file);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_MISSING_FILE);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_BOOKMARKS);
	mhash_write(fd, state->bookmarks);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_BOOKMARKS);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_INDEXED_PAGES);
	mhash_write(fd, state->indexed_pages);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_INDEXED_PAGES);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_EXTENSIONS);
	mhash_write(fd, state->extension);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_EXTENSIONS);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_VISITS);
	mhash_write(fd, state->visits);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_VISITS);

	gzprintf(fd, "<%s>\n", M_STATE_WEB_VIEWS);
	mhash_write(fd, state->views);
	gzprintf(fd, "</%s>\n", M_STATE_WEB_VIEWS);

	gzprintf(fd, "<%s>\n", M_STATE_WEB_HOURS);
	for (i = 0; i < 24; i++) {
		gzprintf(fd, "<hour index=\"%d\">\n", i);
		
		
#define WR(x) \
	gzprintf(fd, "<%s>%ld</%s>", #x, state->hours[i].x, #x);
		
		WR(hits);
		WR(files);
		WR(pages);
		WR(visits);
		WR(hosts);
#undef WR
#define WR(x) \
	gzprintf(fd, "<%s>%.0f</%s>", #x, state->hours[i].x, #x);
		
		WR(xfersize);
#undef WR
		gzprintf(fd, "</hour>\n");
	}
	gzprintf(fd, "</%s>\n", M_STATE_WEB_HOURS);
	
	gzprintf(fd, "<%s>\n", M_STATE_WEB_DAYS);
	for (i = 0; i < 31; i++) {
		gzprintf(fd, "<day index=\"%d\">\n", i);
#define WR(x) \
	gzprintf(fd, "<%s>%ld</%s>", #x, state->days[i].x, #x);
		
		WR(hits);
		WR(files);
		WR(pages);
		WR(visits);
		WR(hosts);
#undef WR
#define WR(x) \
	gzprintf(fd, "<%s>%.0f</%s>", #x, state->days[i].x, #x);
		
		WR(xfersize);
#undef WR
		gzprintf(fd, "</day>\n");
	}
	gzprintf(fd, "</%s>\n", M_STATE_WEB_DAYS);
	gzprintf(fd, "</%s>\n", M_STATE_SECTION_WEB);

	return 0;
}

int mstate_write_mail (mconfig *conf, mstate_mail *state, gzFile *fd) {
	int i;
	
	gzprintf(fd, "<%s>\n", M_STATE_SECTION_MAIL);

	gzprintf(fd, "<%s>\n", M_STATE_MAIL_SENDER);
	mhash_write(fd, state->sender);
	gzprintf(fd, "</%s>\n", M_STATE_MAIL_SENDER);

	gzprintf(fd, "<%s>\n", M_STATE_MAIL_RECEIPIENT);
	mhash_write(fd, state->receipient);
	gzprintf(fd, "</%s>\n", M_STATE_MAIL_RECEIPIENT);
	
	gzprintf(fd, "<%s>\n", M_STATE_MAIL_HOURS);
	for (i = 0; i < 24; i++) {
		gzprintf(fd, "<hour index=\"%d\">\n", i);
		
#define WR(x) \
	gzprintf(fd, "<%s>%ld</%s>\n", #x, state->hours[i].x, #x);
		
		WR(incomming_mails);
		WR(outgoing_mails);
		WR(incomming_bytes);
		WR(outgoing_bytes);
#undef WR
		gzprintf(fd, "</hour>\n");
	}
	gzprintf(fd, "</%s>\n", M_STATE_MAIL_HOURS);

	gzprintf(fd, "<%s>\n", M_STATE_MAIL_DAYS);
	for (i = 0; i < 31; i++) {
		gzprintf(fd, "<day index=\"%d\">\n", i);
#define WR(x) \
	gzprintf(fd, "<%s>%ld</%s>\n", #x, state->days[i].x, #x);
		
		WR(incomming_mails);
		WR(outgoing_mails);
		WR(incomming_bytes);
		WR(outgoing_bytes);
#undef WR
		gzprintf(fd, "</day>\n");
	}
	gzprintf(fd, "</%s>\n", M_STATE_MAIL_DAYS);
	
	gzprintf(fd, "</%s>\n", M_STATE_SECTION_MAIL);

	return 0;
}

int mstate_write_telecom (mconfig *conf, mstate_telecom *state, gzFile *fd) {
	return 0;
}

int mstate_write_traffic (mconfig *conf, mstate_traffic *state, gzFile *fd) {
	return 0;
}

int mstate_write (mconfig *conf, mstate *state, int _add_month, char *subpath) {
	char filename[255], buf[255];
	gzFile *fd;
	
	if (state == NULL) return -1;

	if (subpath) {
		sprintf(filename, "%s/%s/",
			conf->outputdir ? conf->outputdir : ".",
			subpath);
		mkdir(filename, 0755);
		
		if (_add_month == M_STATE_WRITE_DEFAULT) {
			sprintf(filename, "%s/%s/mla.state.xml", 
				conf->outputdir ? conf->outputdir : ".", 
				subpath);
		} else {
			sprintf(filename, "%s/%s/mla.state.xml.%04d%02d", 
				conf->outputdir ? conf->outputdir : ".", 
				subpath,
				state->year, state->month);
		}
	} else {
		if (_add_month == M_STATE_WRITE_DEFAULT) {
			sprintf(filename, "%s/mla.state.xml", 
				conf->outputdir ? conf->outputdir : ".");
		} else {
			sprintf(filename, "%s/mla.state.xml.%04d%02d", 
				conf->outputdir ? conf->outputdir : ".", 
				state->year, state->month);
		}
	}

	
	if (conf->debug_level > 1) {
		fprintf(stderr, "%s.%d: STATE-Filename: %s\n", __FILE__, __LINE__, filename);
	}
	
	if ((fd = gzopen(filename, "wb")) == NULL) {
		fprintf(stderr, "%s.%d: can't open %s: %s\n", __FILE__, __LINE__, filename, strerror(errno));
		return -1;
	}

/* header */
	gzprintf(fd, "<?xml version=\"1.0\"?>\n");
	gzprintf(fd, "<!DOCTYPE state PUBLIC \"http://www.kneschke.de/projekte/modlogan/mla.state.dtd\" \"mla.state.dtd\">\n");
	gzprintf(fd, "<state version=\"%s\" package=\"%s\">\n",
		STATE_FILE_VERSION, PACKAGE);
	
/* global values */
	
	gzprintf(fd, "<global type=\"%s\">", 
		state->week ? "weekly" : "monthly");
	
	gzprintf(fd, "<timestamp>%ld</timestamp>", state->timestamp);
	gzprintf(fd, "<year>%d</year>", state->year);
	gzprintf(fd, "<month>%d</month>", state->month);
	gzprintf(fd, "<week>%d</week>", state->week);
	gzprintf(fd, "</global>\n");

	/* FIXME: input, processor, output */
	if (state->ext) {
		switch(state->ext_type) {
		case M_STATE_TYPE_WEB:
			mstate_write_web(conf, state->ext, fd);
			break;
		case M_STATE_TYPE_TELECOM:
			mstate_write_telecom(conf, state->ext, fd);
			break;
		case M_STATE_TYPE_TRAFFIC:
			mstate_write_traffic(conf, state->ext, fd);
			break;
		case M_STATE_TYPE_MAIL:
			mstate_write_mail(conf, state->ext, fd);
			break;
		default:
			fprintf(stderr, "%s.%d: unknown substate type\n", __FILE__, __LINE__);
		}
	} else {
		if (conf->debug_level > 1)
			fprintf(stderr, "%s.%d: no state extension\n", __FILE__, __LINE__);
	}
	
	gzprintf(fd, "</state>\n");
	gzclose(fd);
	
	return 0;
}

#define M_STATE_DATATYPE_UNSET	0
#define M_STATE_DATATYPE_HASH	1
#define M_STATE_DATATYPE_LONG	2
#define M_STATE_DATATYPE_LIST	3
#define M_STATE_DATATYPE_WEB_ARRAY 4

#define M_STATE_DATASUBTYPE_UNSET	0

int mstate_insert_value(void *dest, int type, xmlNodePtr node) {
	switch (type) {
	case M_STATE_DATATYPE_LONG: {
		int str = *(int *)(dest);
		char *buffer;
			
		if ((buffer = xmlNodeGetContent(node)) != NULL) {
			
			str = strtol(buffer, NULL, 10);
			
			*(int *)(dest) = str;
			
//			fprintf(stderr, "%s.%d: xml-read: %d\n", __FILE__, __LINE__, str);
			
			free(buffer);
		} else {
			fprintf(stderr, "%s.%d: xml-read failed\n", __FILE__, __LINE__);
		}
			
		break;
	}
	
	case M_STATE_DATATYPE_HASH: {
		mdata *data;
		mhash *hash = *(mhash **) dest;
		
		while (node) {
			if (node->type != XML_ELEMENT_NODE) {
				node = node->next;
				continue;
			}
			
			data = mdata_init();
#ifdef DEBUG_DATATYPES
			fprintf(stderr, "%s.%d: got %d\n", __FILE__, __LINE__, data->id);
#endif
			mdata_read(data, node);
		
			if (data->key) {
				mhash_insert(hash, data);
			}
			
			node = node->next;
		}
		
		break;
	}
		
	case M_STATE_DATATYPE_LIST: {
		mdata *data;
		mlist *list = *(mlist **) dest;
		
		while (node) {
			if (node->type != XML_ELEMENT_NODE) {
				node = node->next;
				continue;
			}
			
			data = mdata_init();
#ifdef DEBUG_DATATYPES
			fprintf(stderr, "%s.%d: got %d\n", __FILE__, __LINE__, data->id);
#endif 
			mdata_read(data, node);
		
			if (data->key) {
				mlist_insert(list, data);
			}
			
			node = node->next;
		}
		
		break;
	}
	case M_STATE_DATATYPE_WEB_ARRAY: {	
		long hits, files, pages, visits, hosts;
		double xfersize;
		marray_web *wa;
		const mdata_values data_values[] = {
			{ "hits", M_DATA_FIELDTYPE_LONG, 	&(hits) },
			{ "files", M_DATA_FIELDTYPE_LONG, 	&(files) },
			{ "pages", M_DATA_FIELDTYPE_LONG, 	&(pages) },
			{ "visits", M_DATA_FIELDTYPE_LONG, 	&(visits) },
			{ "hosts", M_DATA_FIELDTYPE_LONG, 	&(hosts) },
			{ "xfersize", M_DATA_FIELDTYPE_DOUBLE, 	&(xfersize) },
			
			{ NULL, M_DATA_FIELDTYPE_UNSET,  NULL}
		};
		
		wa = (marray_web *) dest;
		
		while (node) {
			xmlNodePtr subnode;
			xmlChar *index;
			int i;
			if (node->type != XML_ELEMENT_NODE) {
				node = node->next;
				continue;
			}
			
			subnode = node->children;
			
			while (subnode) {
				if (subnode->type != XML_ELEMENT_NODE) {
					subnode = subnode->next;
					continue;
				}
				
//		fprintf(stderr, "%s.%d: processing %s\n", __FILE__, __LINE__, subnode->name);
				
				i = 0;
				while (data_values[i].string) {
					if (!strcmp(data_values[i].string, subnode->name))
						break;
					i++;
				}
			
				if (data_values[i].string != NULL) {
					if (mdata_insert_value(data_values[i].dest, data_values[i].type, subnode->children)) {
						fprintf(stderr, "%s.%d: insert failed for '%s'\n", __FILE__, __LINE__, subnode->name);
					}
				} else {
					fprintf(stderr, "%s.%d: insert failed, element not found '%s'\n", __FILE__, __LINE__, subnode->name);
				}
				
				subnode = subnode->next;
			}
			
			index = xmlGetProp(node, "index");
			
			if (index) {
				i = strtol(index, NULL, 10);
			
				xmlFree(index);
				
				wa[i].hits = hits;
				wa[i].files = files;
				wa[i].pages = pages;
				wa[i].visits = visits;
				wa[i].hosts = hosts;
				wa[i].xfersize = xfersize;
			} else {
				fprintf(stderr, "%s.%d: index is missing\n", __FILE__, __LINE__);
				return -1;
			}
			
			node = node->next;
		}
		break;
	}

	default: 
		fprintf(stderr, "%s.%d: (insert) unknown type: %d\n", __FILE__, __LINE__, type);
		return -1;
	}
	return 0;
}

typedef struct {
	char	*string;
	int	type;
	void	*dest;
} mstate_values;

int mstate_read_web (mconfig *conf, mstate_web *state, xmlNodePtr node) {
	int i = 0;

	const mstate_values state_values[] = {
		{ M_STATE_WEB_REQ_URL,	M_STATE_DATATYPE_HASH,		&(state->req_url_hash) },
		{ M_STATE_WEB_REQ_METHOD, M_STATE_DATATYPE_HASH, 	&(state->req_meth_hash) },
		{ M_STATE_WEB_REQ_PROTOCOL, M_STATE_DATATYPE_HASH, 	&(state->req_prot_hash) },
		{ M_STATE_WEB_REF_URL,	M_STATE_DATATYPE_HASH,		&(state->ref_url_hash) },
		{ M_STATE_WEB_OS,	M_STATE_DATATYPE_HASH,		&(state->os_hash) },
		{ M_STATE_WEB_USERAGENT, M_STATE_DATATYPE_HASH,	&(state->ua_hash) },
		{ M_STATE_WEB_HOST,	M_STATE_DATATYPE_HASH,		&(state->host_hash) },
		{ M_STATE_WEB_STATUS,	M_STATE_DATATYPE_HASH,		&(state->status_hash) },
		{ M_STATE_WEB_VISIT,	M_STATE_DATATYPE_LIST,		&(state->visit_list) },
		{ M_STATE_WEB_HOURS,	M_STATE_DATATYPE_WEB_ARRAY, 	&(state->hours) },
		{ M_STATE_WEB_DAYS,	M_STATE_DATATYPE_WEB_ARRAY, 	&(state->days) },
		{ M_STATE_WEB_COUNTRIES, M_STATE_DATATYPE_HASH,	&(state->country_hash) },
		{ M_STATE_WEB_ROBOTS,	M_STATE_DATATYPE_HASH,		&(state->robots) },
		{ M_STATE_WEB_SEARCHSITE, M_STATE_DATATYPE_HASH, 	&(state->searchsite) },
		{ M_STATE_WEB_SEARCHSTRING, M_STATE_DATATYPE_HASH, 	&(state->searchstring) },
		{ M_STATE_WEB_INT_ERROR, M_STATE_DATATYPE_HASH,		&(state->status_internal_error) },
		{ M_STATE_WEB_MISSING_FILE, M_STATE_DATATYPE_HASH,  &(state->status_missing_file) },
		{ M_STATE_WEB_BOOKMARKS, M_STATE_DATATYPE_HASH,		&(state->bookmarks) },
		{ M_STATE_WEB_INDEXED_PAGES, M_STATE_DATATYPE_HASH, 	&(state->indexed_pages) },
		{ M_STATE_WEB_EXTENSIONS, M_STATE_DATATYPE_HASH, 	&(state->extension) },
		{ M_STATE_WEB_VISITS,    M_STATE_DATATYPE_HASH, 	&(state->visits) },
		{ M_STATE_WEB_VIEWS,    M_STATE_DATATYPE_HASH, 		&(state->views) },
		
		{ NULL, M_STATE_DATATYPE_UNSET,  NULL}
	};

	while (node) {
		if (node->type != XML_ELEMENT_NODE) {
			node = node->next;
			continue;
		}
		
		//fprintf(stderr, "%s.%d: processing %s\n", __FILE__, __LINE__, node->name);
		
		i = 0;
		while (state_values[i].string) {
			if (!strcmp(state_values[i].string, node->name))
				break;
			i++;
		}

		if (state_values[i].string != NULL) {
			if (mstate_insert_value(state_values[i].dest, state_values[i].type, node->children)) {
				fprintf(stderr, "%s.%d: insert failed for '%s'\n", __FILE__, __LINE__, node->name);
			}
		} else {
			fprintf(stderr, "%s.%d: insert failed, element not found '%s'\n", __FILE__, __LINE__, node->name);
		}
		
		node = node->next;
	}
	
	return 0;
}

int mstate_read_traffic (mconfig *conf, mstate_traffic *state, xmlNodePtr doc) {
	int i = 0;
#if 0
	const mstate_values state_values[] = {

		{ NULL, M_STATE_DATATYPE_UNSET,  NULL}
	};

	while (state_values[i].string) {
		if (!strcmp(state_values[i].string, key))
			break;
		i++;
	}

	if (!state_values[i].string) return -1;
	
	if (mstate_insert_value(state_values[i].dest, state_values[i].type, state_values[i].subtype, f)) {
		return -1;
	}
#endif
	return 0;
}

int mstate_read_telecom (mconfig *conf, mstate_telecom *state, xmlNodePtr doc) {
	int i = 0;
#if 0
	const mstate_values state_values[] = {

		{ NULL, M_STATE_DATATYPE_UNSET,  NULL}
	};

	while (state_values[i].string) {
		if (!strcmp(state_values[i].string, key))
			break;
		i++;
	}

	if (!state_values[i].string) return -1;
	
	if (mstate_insert_value(state_values[i].dest, state_values[i].type, state_values[i].subtype, f)) {
		return -1;
	}
#endif
	return 0;
}

int mstate_read_global (mconfig *conf, mstate *state, xmlNodePtr node) {
	int i;
	const mstate_values state_values[] = {
		{ "year", M_STATE_DATATYPE_LONG, 	&(state->year) },
		{ "month", M_STATE_DATATYPE_LONG, 	&(state->month) },
		{ "week", M_STATE_DATATYPE_LONG, 	&(state->week) },
		{ "timestamp", M_STATE_DATATYPE_LONG,  &(state->timestamp) },
		
		{ NULL, M_STATE_DATATYPE_UNSET,  NULL}
	};
	
	while (node) {
		if (node->type != XML_ELEMENT_NODE) {
			node = node->next;
			continue;
		}
		
		i = 0;
		while (state_values[i].string) {
			if (!strcmp(state_values[i].string, node->name))
				break;
			i++;
		}

		if (state_values[i].string != NULL) {
			if (mstate_insert_value(state_values[i].dest, state_values[i].type, node)) {
				fprintf(stderr, "%s.%d: insert failed for '%s'\n", __FILE__, __LINE__, node->name);
			}
		} else {
			fprintf(stderr, "%s.%d: insert failed, element not found '%s'\n", __FILE__, __LINE__, node->name);
		}
		
		node = node->next;
	}
	
	return 0;
}

int mstate_read (mconfig *conf, mstate *state, int year, int month, char *subpath) {
	char filename[255];
	
	xmlDocPtr doc = NULL;
	xmlNodePtr node = NULL;
	
	if (year == 0 || month == 0) {
		sprintf(filename, "%s/%s/mla.state.xml", 
			conf->outputdir, 
			subpath ? subpath : "");
	} else {
		sprintf(filename, "%s/%s/mla.state.xml.%04d%02d", 
			conf->outputdir, 
			subpath ? subpath : "",
			year, month );
	}
	//fprintf(stderr, "%s.%d: mstate_read: xmlParseFile\n", __FILE__, __LINE__);
	if ((doc = xmlParseFile(filename)) == NULL) {
		fprintf(stderr, "%s.%d: *grmpf* can't parse xml file '%s': %s\n", __FILE__, __LINE__, filename, strerror(errno));
		return -1;
	}
	//fprintf(stderr, "%s.%d: mstate_read: xmlParseFile: done\n", __FILE__, __LINE__);
//	xmlDebugDumpDocument(NULL, doc);


	/* the first node is the documentroot */
	node = xmlDocGetRootElement(doc);
	
	/* get the sections */
	node = node->children;
	
	while (node) {
		/* at this stage we should get all the master sections */
		//fprintf(stderr, "%s.%d\n", __FILE__, __LINE__);
		//xmlDebugDumpNode(stderr, node, 0);
		
		if (node->type == XML_TEXT_NODE) {
			node = node->next;
			continue;
		}
		
		/* mapping section name to parser */
		if (strcmp(node->name, M_STATE_SECTION_WEB) == 0) {
			if (state->ext) {
				if (state->ext_type != M_STATE_TYPE_WEB) {
					xmlFreeDoc(doc);
					return -1;
				}
			} else {
				state->ext = mstate_init_web();
				state->ext_type = M_STATE_TYPE_WEB;
			}
				
			mstate_read_web(conf, state->ext, node->children);
		} else if (strcmp(node->name, M_STATE_SECTION_TELECOM) == 0) {
			if (state->ext) {
				if (state->ext_type != M_STATE_TYPE_TELECOM) {
					xmlFreeDoc(doc);
					return -1;
				}
			} else {
				state->ext = mstate_init_telecom();
				state->ext_type = M_STATE_TYPE_TELECOM;
			}
				
			mstate_read_telecom(conf, state->ext, node->children);
		} else if (strcmp(node->name, M_STATE_SECTION_TRAFFIC) == 0) {
			if (state->ext) {
				if (state->ext_type != M_STATE_TYPE_TRAFFIC) {
					xmlFreeDoc(doc);
					return -1;
				}
			} else {
				state->ext = mstate_init_traffic();
				state->ext_type = M_STATE_TYPE_TRAFFIC;
			}
				
			mstate_read_traffic(conf, state->ext, node->children);
		} else if (strcmp(node->name, M_STATE_SECTION_GLOBAL) == 0) {
			mstate_read_global(conf, state, node->children);
		} else {
			fprintf(stderr, "%s.%d: no state-file parser for %s\n", __FILE__, __LINE__, node->name);
		}

		node = node->next;
	}

	xmlFreeDoc(doc);
	
	return 0;
}
